""" A Python Class for Non Oriented Graphs
LG nov 2016 and Feb 2020 - updated June 2021
adapted from http://www.python-course.eu/graphs_python.php
"""

# for dot output
from graphviz import Digraph

debug = False


class Error(Exception):
    """Base class for exceptions in this module."""
    pass


class GraphError(Error):
    """Exception raised for self loops
    """
    def __init__(self, message):
        self.message = message


class Graph(object):

    def __init__(self, graph_dict=None):
        """ initializes a graph object
            If no dictionary or None is given,
            an empty dictionary will be used
            We could also check consistancy here.
        """
        if graph_dict is None:
            graph_dict = {}
        self.__graph_dict = graph_dict
        assert self.verify_consistancy()

    def vertices(self):
        """ returns the vertices of a graph """
        return list(self.__graph_dict.keys())

    def edges(self):
        """ returns the edges of a graph """
        return self.__generate_edges()

    def neighbours(self, vertex):
        return self.__graph_dict[vertex]

    def add_vertex(self, vertex):
        """ If the vertex "vertex" is not in
            self.__graph_dict, a key "vertex" with an empty
            list as a value is added to the dictionary.
            Otherwise nothing has to be done.
        """
        if vertex not in self.__graph_dict:
            self.__graph_dict[vertex] = []

    def add_edge(self, edge):
        """ edge should be a pair and not (c,c)
        """
        try:
            (vertex1, vertex2) = edge
            if vertex1 == vertex2:
                raise GraphError("no self loop")
            if vertex1 in self.__graph_dict:
                self.__graph_dict[vertex1].append(vertex2)
            else:
                self.__graph_dict[vertex1] = [vertex2]
            if vertex2 in self.__graph_dict:
                self.__graph_dict[vertex2].append(vertex1)
            else:
                self.__graph_dict[vertex2] = [vertex1]
        except GraphError as s:
            print("pb with adding edge: "+s.message)
            pass

    def __generate_edges(self):
        """ A static method generating the set of edges
        (they appear twice in the dictionnary). Returns a list of sets.
        """
        edges = []
        for vertex in self.__graph_dict:
            for neighbour in self.__graph_dict[vertex]:
                if {neighbour, vertex} not in edges:
                    edges.append({vertex, neighbour})
        return edges

    def __str__(self):
        res = "vertices: "
        for k in self.__graph_dict:
            res += str(k) + " "
        res += "\nedges: "
        for edge in self.__generate_edges():
            res += str(edge) + " "
        return res

    def verify_consistancy(self):
        """
        Verify the consistancy of the graph, ie : if e1->e2 then e2->e1
        """
        gdict = self.__graph_dict
        for v in gdict.keys():
            for neighbour in gdict[v]:
                if v not in gdict[neighbour]:
                    return False
        return True

    def delete_vertex(self, vertex):
        """ delete vertex and all the adjacent edges"""
        gdict = self.__graph_dict
        for neighbour in gdict[vertex]:
            gdict[neighbour].remove(vertex)
        del gdict[vertex]

    def delete_edge(self, edge):
        (v1, v2) = edge
        self.__graph_dict[v1].remove(v2)
        self.__graph_dict[v2].remove(v1)

    def dfs_traversal(self, root):  # V2 slides 26/37
        """
        Deep first search traversal for graphs
        Imperative version with a stack.
        Warning : the complexity is not minimal.
        """
        seen = []  # L
        todo = [root]
        gdict = self.__graph_dict
        # TODO : Implement DFS : use .pop()
        return seen

    def is_reachable_from(self, v1, v2):
        # Could be better implemented
        return v2 in self.dfs_traversal(v1)

    def dfs_traversal_v2(self, root, seen_d={}):
        """
        Deep first search traversal for graphs with complexity n+m
        Imperative version with a stack.
        Use a dict to store the "seen" property
        The 3rd argument is used to propagate the
        "seen" information for the full graph.
        """
        gdict = self.__graph_dict
        if not seen_d:
            seen_d = {k: False for k in gdict.keys()}
        seen_l = []  # tracks nodes that are reachable from root
        todo = [root]
        # TODO : finish!
        return seen_l, seen_d

    def full_dfs_traversal(self):
        """
         Deep first search full traversal for graphs
         Use the previous function from each "non seen" node of the graph
        """
        allseen = []
        gdict = self.__graph_dict
        seen_d = {k: False for k in gdict.keys()}
        for root in gdict.keys():
            if not(seen_d[root]):
                seen_l,seen_d= self.dfs_traversal_v2(root,seen_d)
                allseen.append(seen_l)
        return allseen

    def bfs_traversal(self, root):  # slide 26/36 du cours généralités
        """
        Breadth first search traversal for graphs
        Imperative version with a queue
        Warning : the complexity is not O(m+n)!
        """
        seen = []  # == L
        todo = [root]  # B
        gdict = self.__graph_dict
        # TODO : Implement BFS : use .pop(0) to dequeue
        return seen  # L

    def has_a_cycle_aux(self, root, level_d={}):
        """
        Search for a cycle in the connex component that contains root
        Use levels information (variant of BFS) - a simplier algorithm based
        on DFS can be found here
        https://www.iro.umontreal.ca/~hamelsyl/parcoursA09.pdf
        If there is a u with a neighbour w such that level[w] >= level[u],
        there is a cycle in the graph.
        """
        gdict = self.__graph_dict
        if not level_d:
             level_d = {k: None for k in gdict.keys()}
        level_d[root] = 0
        todo = [root]
        parent = {k: None for k in gdict.keys()}
        return False, [], level_d

    def has_a_cycle(self):
        """
        Search for a cycle in the full (possibly non connex) graph
        returns true/false + a list depicting a cycle if any
        """
        # allseen = []
        gdict = self.__graph_dict
        level_d = {k: None for k in gdict.keys()}
        for root in gdict.keys():
            if level_d[root] is None:  # search in CC from root.
                if debug:
                    print("call to has a cycle_aux with root=", root)
                b, cycle, level_d = self.has_a_cycle_aux(root, level_d)
                if b:
                    return (True, cycle)
        return (False, [])

    def connex_components(self):
        """
        Naive version of connex_components
        the full_dfs_traversal gives a less naive version
        """
        components = []
        todo = self.vertices()
        done = []
        # TODO : find CCs
        return components

    def print_dot(self, name="toto", colors={}):
        """
        Use graphviz to print the graph. Colors is optional.
        works with string nodes as well as int nodes.
        """
        color_names = ['red', 'blue', 'green', 'yellow', 'cyan', 'magenta'] + [f"grey{i}" for i in range(0, 100, 10)]
        color_shapes = ['ellipse', 'polygon', 'box', 'circle', 'egg', 'pentagon', 'hexagon']
        dot = Digraph(comment=name, format='png') # change into pdf if you want.
        for k in self.__graph_dict:
            shape = None
            if not colors:
                color = "red"
            else:
                n = colors[k]
                if n < len(color_names):
                    color = color_names[colors[k]]
                else:
                    color = "black" # Too many colors anyway, it won't be readable.
                shape = color_shapes[n % len(color_shapes)]
            dot.node(str(k), color=color, shape=shape)
        for edge in self.__generate_edges():
            # print(edge)
            (v1, v2) = list(edge)[0], list(edge)[1]
            dot.edge(str(v1), str(v2), dir="none")
        if debug:
            print(dot.source)
        dot.render(name, cleanup=True)

    def color(self):
        """Color the graph with an unlimited number of colors.

        Return a dict vertice -> color, where color is an integer (0, 1,
        ...)."""
        coloring, _, _ = self.color_with_k_colors()
        return coloring

    # see algo of the course (based on the Kempe algo)
    def color_with_k_colors(self, K=None, avoidingnodes=()):
        """color with <= K color (if K is unspecified, use unlimited colors).

        Return 3 values:
        - a map vertex-> color
        - a Boolean, True if the coloring succeedeed
        - The set of nodes actually colored

        Do not color vertices belonging to avoidingnodes.

        Continue even if the algo fails.
        """
        if K is None:
            K = len(self.__graph_dict)
        todo_vertices = []
        is_total = True
        gcopy = Graph(self.__graph_dict.copy())
        # suppress nodes that are not to be considered.
        for node in avoidingnodes:
            gcopy.delete_vertex(node)
        # append nodes in the list according to their degree and node number:
        while gcopy.__graph_dict:
            todo = list(gcopy.__graph_dict)
            todo.sort(key=lambda v: (len(gcopy.__graph_dict[v]), str(v)))
            lower = todo[0]
            todo_vertices.append(lower)
            gcopy.delete_vertex(lower)
        # Now reverse the list: first elements are those with higher degree
        # print(todo_vertices)
        todo_vertices.reverse()  # in place reversal
        # print(todo_vertices)
        coloring = {}
        colored_nodes = []
        # gdict will be the coloring map to return
        gdict = self.__graph_dict
        for v in todo_vertices:
            seen_neighbours = [x for x in gdict[v] if x in coloring]
            choose_among = [i for i in range(K) if not(
                i in [coloring[v1] for v1 in seen_neighbours])]
            if choose_among:
                # if the node can be colored, I choose the minimal color.
                color = min(choose_among)
                coloring[v] = color
                colored_nodes.append(v)
            else:
                # if I cannot color some node, the coloring is not Total
                # but I continue
                is_total = False
        return (coloring, is_total, colored_nodes)
